// This is an EXTREMELY small program that will shrink or expand the lcd panel image.
// Phillip Lee Hellewel, 2004-Jan-17.
//
// Thanks to T. N. Roberts for his s3switch program,
// which inspired me to write this (ok, and I borrowed some code too).

// Linux x86 only.

#include <stdio.h>
#include "lrmi.h"
#define extern
#include <asm/io.h>
#undef extern

// Usage: s3lcd [stretch|shrink]

// Define the Savage chip classes.  PCI id's stolen from xf86PciInfo.h

#define PCI_CHIP_SAVAGE3D	0x8A20
#define PCI_CHIP_SAVAGE3D_MV	0x8A21
#define PCI_CHIP_SAVAGE4	0x8A22
#define PCI_CHIP_SAVAGE2000	0x9102
#define PCI_CHIP_PROSAVAGE_PM	0x8A25
#define PCI_CHIP_PROSAVAGE_KM	0x8A26
#define PCI_CHIP_SAVAGE_MX_MV	0x8c10
#define PCI_CHIP_SAVAGE_MX	0x8c11
#define PCI_CHIP_SAVAGE_IX_MV	0x8c12
#define PCI_CHIP_SAVAGE_IX	0x8c13
#define PCI_CHIP_TWISTERP	0x8d01
#define PCI_CHIP_TWISTERK	0x8d02
#define PCI_CHIP_PROSAVAGE_DDR	0x8d03
#define PCI_CHIP_PROSAVAGE_DDRK	0x8d04
#define PCI_CHIP_SUPSAV_MX128	0x8c22
#define PCI_CHIP_SUPSAV_MX64	0x8c24
#define PCI_CHIP_SUPSAV_MX64C	0x8c26
#define PCI_CHIP_SUPSAV_IX128SDR	0x8c2a
#define PCI_CHIP_SUPSAV_IX128DDR	0x8c2b
#define PCI_CHIP_SUPSAV_IX64SDR	0x8c2c
#define PCI_CHIP_SUPSAV_IX64DDR	0x8c2d
#define PCI_CHIP_SUPSAV_IXCSDR	0x8c2e
#define PCI_CHIP_SUPSAV_IXCDDR	0x8c2f

void
usage()
{
    puts("Usage: s3lcd [shrink|stretch]" );
    puts(" Use stretch to expand the LCD panel to the full screen.");
    puts(" Use shrink to restore it again.");
}

void
IOAccess( int enable )
{
    /* Allow or disallow access to I/O ports. */

    ioperm( 0x40, 4, enable );
    ioperm( 0x61, 1, enable );
    ioperm( 0x80, 1, enable );
    ioperm( 0x3b0, 0x30, enable );
}

void
fetch_bios_data()
{
	unsigned int PCIid = 0;

    // Figure out what kind of Savage it is.
    outb( 0x2d, 0x3d4 );
    PCIid = inb( 0x3d5 ) << 8;
    outb( 0x2e, 0x3d4 );
    PCIid |= inb( 0x3d5 );

    switch( PCIid ) {
        case PCI_CHIP_SAVAGE3D:
        case PCI_CHIP_SAVAGE3D_MV:
        case PCI_CHIP_SAVAGE4:
        case PCI_CHIP_SAVAGE2000:
        case PCI_CHIP_PROSAVAGE_PM:
        case PCI_CHIP_PROSAVAGE_KM:
        case PCI_CHIP_TWISTERP:
        case PCI_CHIP_TWISTERK:
        case PCI_CHIP_PROSAVAGE_DDR:
        case PCI_CHIP_PROSAVAGE_DDRK:
        case PCI_CHIP_SAVAGE_MX_MV:
        case PCI_CHIP_SAVAGE_MX:
        case PCI_CHIP_SAVAGE_IX_MV:
        case PCI_CHIP_SAVAGE_IX:
        case PCI_CHIP_SUPSAV_MX128:
        case PCI_CHIP_SUPSAV_MX64:
        case PCI_CHIP_SUPSAV_MX64C:
        case PCI_CHIP_SUPSAV_IX128SDR:
        case PCI_CHIP_SUPSAV_IX128DDR:
        case PCI_CHIP_SUPSAV_IX64SDR:
        case PCI_CHIP_SUPSAV_IX64DDR:
        case PCI_CHIP_SUPSAV_IXCSDR:
        case PCI_CHIP_SUPSAV_IXCDDR:
            break;
        default:
            printf( "PCI id is not a recognized Savage: %04x\n", PCIid );
            exit(-1);
    }
}


int
size_lcd(int stretch)
{
    struct LRMI_regs r;

    if( !LRMI_init() )
        return 1;

    memset( &r, 0, sizeof(r) );
    r.eax = 0x4f14;
    r.ebx = 0x306;
    r.ecx = stretch ? 1 : 2;

    if( !LRMI_int(0x10, &r) ) {
        fprintf( stderr, "Could not %s the LCD panel image.\n", stretch ? "stretch" : "shrink" );
        return 1;
    }

    return 0;
}

int
main(int argc, char* argv[])
{
    int stretch = -1;

    if( geteuid() != 0 ) {
        fprintf( stderr, "s3lcd must be setuid root.\n" );
        return -1;
    }

    // Let's see what the user wants to do.
    if( argc < 2 || strcmp(argv[1], "--help") == 0 ) {
        usage();
        return 0;
    } else if( strcasecmp(argv[1], "stretch") == 0 )
        stretch = 1;
    else if( strcasecmp(argv[1], "shrink" ) == 0 )
        stretch = 0;
    else {
        fprintf( stderr, "Unknown argument: %s\n", argv[1] );
        usage();
        return -1;
    }

    IOAccess( 1 );

    fetch_bios_data();

    size_lcd(stretch);

    IOAccess( 0 );

    return 0;
}
